#!/bin/bash
mkdir -p outputs/
touch "$(pwd)/filebrowser.db"

# stdout/stderr directed to files for debugging.
#exec > >(tee revd.out)
#exec 2> revd.err

# function for getting the last folder in a path
# we're not reading from tapisjob.env since that is not always present for all fork + batch jobs
JOB_ID=`basename "$(pwd)"`
echo "Job id: $JOB_ID" >> debug.txt

PID_FILE="$(pwd)/${JOB_ID}.pid"

CURRENT_SCRIPT_DIR=$(dirname "$0")
REVSSH_SCRIPT_PATH="$CURRENT_SCRIPT_DIR/revssh/revssh-client-v5.sh"
CLEANUP_SCRIPT_PATH="$CURRENT_SCRIPT_DIR/osp_launcher.clean.sh"

echo "PID_FILE: $PID_FILE" >> debug.txt
echo "CURRENT_SCRIPT_DIR: $CURRENT_SCRIPT_DIR" >> debug.txt
echo "REVSSH_SCRIPT_PATH: $REVSSH_SCRIPT_PATH" >> debug.txt
echo "CLEANUP_SCRIPT_PATH: $CLEANUP_SCRIPT_PATH" >> debug.txt

# Function to build a command array with environment variables after parsing them from the input
build_cmd_with_env_vars() {
  # Store the input command as an array
  local input_cmd=("$@")
  local -a env_vars
  local -a prog_and_args
  local key_value
  local processing_envs=0
  local osp_placeholder_index=-1

  # Iterate over the command array to process each item
  for i in "${!input_cmd[@]}"; do
    if [[ "${input_cmd[$i]}" == "--env" ]]; then
      processing_envs=1
      # Split the environment variables string into an array
      IFS=',' read -ra key_value <<< "${input_cmd[$((i + 1))]}"
      for env_var in "${key_value[@]}"; do
        # Export each environment variable
        IFS='=' read -r key value <<< "$env_var"
        export "$key=$value"
	echo "export $key=$value" >> debug.txt
      done
      # Skip the next item since it's part of the --env option
      ((i++))
    elif [[ "${input_cmd[$i]}" == "osp_placeholder.sif" ]]; then
      osp_placeholder_index=$i
      processing_envs=0
      break # Stop processing once osp_placeholder.sif is found
    fi
  done

  # Return the array starting from the element right after osp_placeholder.sif keyword
  if [ $osp_placeholder_index -ne -1 ]; then
    prog_and_args=("${input_cmd[@]:$((osp_placeholder_index + 1))}")
  fi

  echo "${prog_and_args[@]}"
}

# build_cmd_with_env_vars() {
#   # Store the input command as an array
#   local input_cmd=("$@")

#   # Extract the environment variables from the input command
#   local env_vars=($(echo "${input_cmd[@]}" | grep -oP '(?<=--env ).*?(?= osp_placeholder.sif)' | tr ',' ' '))

#   # Extract the program name and its arguments from the input command
#   local prog_and_args=($(echo "${input_cmd[@]}" | grep -oP '(?<=osp_placeholder.sif ).*'))

#   # Combine the environment variables with the extracted program and its arguments
#   local cmd_with_env_vars=("${env_vars[@]}" "${prog_and_args[@]}")

#   # Return the combined command array
#   echo "${cmd_with_env_vars[@]}"
# }

find_unused_local_port() {
  local i port
  local -A p

  netstat -aln | awk '
    $6 == "LISTEN" {
      if ($4 ~ "[.:][0-9]+$") {
        split($4, a, /[:.]/)
        port = a[length(a)]
        p[port] = 1
      }
    }
    END {
      for (i = 3000; i < 65000 && p[i]; i++){};
      if (i == 65000) {exit 1};
      print i
    }
  '
}

# Function to replace the value for an option with a quoted value
quote_value() {
    local input_string="$1"
    local option="$2"

    echo "$input_string" | awk -v opt="$option" '
    {
        for (i=1; i<=NF; i++) {
            if ($i == opt) {
                # Join the next fields until we hit an option or end of line
                val=""
                for (j=i+1; j<=NF; j++) {
		    if ($j ~ /^--/ || $j == "osp_placeholder.sif") break
                    if (val != "") val = val " " $j
                    else val = $j
                }
                printf "%s \"%s\" ", opt, val
                i = j-1  # Skip ahead
            } else {
                printf "%s ", $i
            }
        }
        print ""
    }'
}

is_executable_app=$1
setup_app_proxying=$2
satellite_host=$3
job_proxy_id=$4

# shift to remove the first four arguments
shift 4

container_cmd="$@"
echo "Container command: $container_cmd" >> ./debug.txt
# Update the string for --bind and --env options
container_cmd=$(quote_value "$container_cmd" "--bind")
container_cmd=$(quote_value "$container_cmd" "--env")
echo "Updated container command: $container_cmd" >> ./debug.txt

# get all the args as a single string (except the first two args)
# container_cmd=("$@")
# Create an array from the input string
IFS=' ' read -r -a container_cmd_array <<< "$container_cmd"

# Reconstruct the array to handle quoted substrings
container_cmd=()
in_quotes=false
temp=""

for element in "${container_cmd_array[@]}"; do
    echo "element: $element" >> debug.txt
    if [[ $in_quotes == true ]]; then
        if [[ $element == *\" ]]; then
            temp+=" ${element%\"}"
            container_cmd+=("$temp")
            temp=""
            in_quotes=false
        else
            temp+=" $element"
        fi
    else
        if [[ $element == \"* ]]; then
            if [[ $element == *\" ]]; then
		element=${element#\"}
		element=${element%\"}
		echo "last: ${element}" >> debug.txt
                container_cmd+=("${element}")
		# container_cmd+=("${element%\"}")
            else
                temp="${element#\"}"
                in_quotes=true
            fi
        else
            container_cmd+=("$element")
        fi
    fi
done

# Print the array elements to verify
printf '%s\n' "${container_cmd[@]}" >> debug.txt

# Define an associative array with keys that require quoting
declare -A keys_to_quote=(
    ["_tapisJobName"]=1
    ["_tapisArchiveSystemDir"]=1
    ["_tapisExecSystemExecDir"]=1
    ["_tapisExecSystemInputDir"]=1
    ["_tapisExecSystemOutputDir"]=1
)

# Modify the container_cmd array to add in the --pid-file option or other options based on conditions
for i in "${!container_cmd[@]}"; do
    echo "index: $i value: ${container_cmd[$i]}" >> debug.txt
    if [[ "${container_cmd[$i]}" == "--bind" ]]; then
	# Split the pair into key and value based on the equals sign
        IFS=':' read -r key value <<< "${container_cmd[$i+1]}"
	key="\"$key\""
	container_cmd[$i+1]="$key:$value"
    fi

    if [[ "${container_cmd[$i]}" == "--env" ]]; then
	# Split the input string into an array based on commas
	IFS=',' read -r -a env_pairs <<< "${container_cmd[$i+1]}"

	# Initialize an empty string to hold the modified pairs
	modified_string=""

	# Iterate through each key-value pair
	for env_pair in "${env_pairs[@]}"; do
    		# Split the pair into key and value based on the equals sign
    		IFS='=' read -r key value <<< "$env_pair"
		# Check if the current key needs quoting
		if [[ ${keys_to_quote["$key"]} ]]; then
    			# Add double quotes around the value
    			value="\"$value\""
		fi

    		# Append the modified pair to the modified string
    		modified_string+="$key=$value,"
	done

	# Remove the trailing comma from the modified string
	modified_string="${modified_string%,}"

	echo "$modified_string" >> debug.txt
	container_cmd[$i+1]="$modified_string"
    fi

    # The following code never worked.
    # if [[ "${container_cmd[$i]}" == "singularity instance start"* ]]; then
    #     container_cmd[$i]="${container_cmd[$i]} --pid-file \"${PID_FILE}\""
    # elif [[ "${container_cmd[$i]}" == "docker run"* ]]; then
    #     container_cmd[$i]="${container_cmd[$i]} --network \"${JOB_ID}\""
    # fi
done

# I'm not clear on why the following code is required.
if [[ "${container_cmd[0]}" == "singularity" && "${container_cmd[1]}" == "instance" && "${container_cmd[2]}" == "start" ]]; then
	container_cmd[2]="${container_cmd[2]} --pid-file \"${PID_FILE}\" "
elif [[ "${container_cmd[0]}" == "docker" && "${container_cmd[1]}" == "run" ]]; then
        container_cmd[1]="${container_cmd[1]} --network \"${JOB_ID}\" "
fi

echo "Container command: ${container_cmd[*]}" >> debug.txt

# Define the OSP_APP_PORT
if [ "$setup_app_proxying" == "1" ]; then
    if [[ "${container_cmd[0]}" == "docker"* ]]; then
        # if we're running a docker container, parse the OSP_APP_PORT from the env vars file tapis generated
        OSP_APP_PORT=$(grep -oP '^OSP_APP_PORT=\K.*$' "$(pwd)/tapisjob.env")
    else
        # otherwise, find an open port
        OSP_APP_PORT=$(find_unused_local_port)
    fi
else
    # OSP_APP_PORT is set to -1 for cases where it's not used (non-web/vnc app, etc.)
    OSP_APP_PORT="-1"
fi

# Replace all instances of '[OSP_APP_PORT]' in the container_cmd array with $OSP_APP_PORT
for i in "${!container_cmd[@]}"; do
    container_cmd[$i]=${container_cmd[$i]/\[OSP_APP_PORT\]/$OSP_APP_PORT}
done

echo "Is executable app: ${is_executable_app}, Setup app proxying: ${setup_app_proxying}, Job proxy id: ${job_proxy_id}, Job ID: ${JOB_ID}, Satellite host: ${satellite_host}, OSP app port: $OSP_APP_PORT" > ./ospjob_proxy_info.txt

# If we need to setup app proxying,
# run revssh-client in the background
if [[ "${container_cmd[0]}" == "docker"* ]]; then
    # create the docker network
    docker network create "${JOB_ID}" > /dev/null 2>&1
fi

if [ "$setup_app_proxying" == "1" ]; then
    echo "running revssh-client in the bg"
    mkdir -p ./.revssh
    mkdir -p ./.revssh/.locks
    
    # auto-create the token file for this initial job proxy id
    touch ./.revssh/${job_proxy_id}.token
    
    # start the reverse ssh tunnel client in the bg
    if [[ "${container_cmd[0]}" == "docker"* ]]; then
        docker run -d --rm --network "${JOB_ID}" \
                   --name="revssh-${JOB_ID}" \
                   -v "$(pwd)/.revssh":/.revssh \
                   -it revssh-client:v5 \
                    $satellite_host "$JOB_ID" $OSP_APP_PORT /.revssh > /dev/null 2>&1
    else
        REVSSH_CLIENT_PID_FILE=./revssh-client.pid
        $REVSSH_SCRIPT_PATH $satellite_host localhost $OSP_APP_PORT ./.revssh > "${JOB_ID}_revssh_client.txt" 2>&1 &
        REVSSH_CLIENT_PID=$!
        echo $REVSSH_CLIENT_PID > $REVSSH_CLIENT_PID_FILE
    fi
fi

# is this an executable app?
if [ "$is_executable_app" == "1" ]; then
    echo "executable app strategy"
    
    # run the executable cmd
    echo "running job container..."
    export OSP_APP_PORT=$OSP_APP_PORT

    executable_app_cmd=($(build_cmd_with_env_vars "${container_cmd[@]}"))
    echo "Executable app command: ${executable_app_cmd[*]}" >> debug.txt
    
    eval "${executable_app_cmd[@]}" > outputs/ospjob.log 2>&1

    # If we setup app proxying, run the cleanup script
    if [ "$setup_app_proxying" == "1" ]; then
        # cleanup
        echo "Cleaning up..."
        
        # This will remove the revssh-client process
        kill $REVSSH_CLIENT_PID
    fi

# check if the container command is a docker command
# by checking if the first word in the command is "docker"
elif [[ "${container_cmd[0]}" == "docker"* ]]; then
    # run the docker command
    CID_FILE="$(pwd)/${JOB_ID}.cid"
    eval "${container_cmd[@]}" > outputs/ospjob.log 2>&1
    
    # create the cid file
    if [ -r "${CID_FILE}" ]; then
        CONTAINER_ID=$(cat "${CID_FILE}")
    else
        # Write container id to cid file
        CONTAINER_ID=$(docker container ls --all --quiet --no-trunc --filter "name=^\"${JOB_ID}\"$")
        echo ${CONTAINER_ID} > "${CID_FILE}"
    fi

    # capture Docker process stdout/stderr log
    nohup docker logs --follow ${CONTAINER_ID} > outputs/ospjob.log 2>&1 &

    # start the cleanup process in the background
    nohup $CLEANUP_SCRIPT_PATH 1 "${JOB_ID}" > "${JOB_ID}_cleanup.txt" 2>&1 &

    # echoing the container id here, since that's what tapis's default launcher script ends up doing
    echo $CONTAINER_ID

# check if the container command is a singularity instance start command
# by checking if the first word in the command is "singularity"
# and the second word is "instance"
# and the third word is "start"
elif [[ "${container_cmd[0]}" == "singularity" && "${container_cmd[1]}" == "instance" && "${container_cmd[2]}" == "start" ]]; then
    echo "singularity instance start strategy" >> debug.txt

    # If we need to setup app proxying,
    # launch the clean up process in the bg
    if [ "$setup_app_proxying" == "1" ]
    then
	    echo "Running clean up script in background..."
	    $CLEANUP_SCRIPT_PATH 0 "${JOB_ID}" "${PID_FILE}" $REVSSH_CLIENT_PID > "${JOB_ID}_cleanup.txt" 2>&1 &
	    CLEANUP_SCRIPT_PID=$!
    fi

    # run this container cmd
    echo "running job container..." >> debug.txt
    echo "container cmd: ${container_cmd[@]}" >> debug.txt
    eval "${container_cmd[@]}" > outputs/ospjob.log 2>&1

    # If we setup app proxying, run the cleanup script
    if [ "$setup_app_proxying" == "1" ]
    then
	    # wait for the cleanup process to exit
	    # We need to do this, cuz otherwise this script will just end right after starting the singularity container in the bg, and the slurm job will be cancelled!
	    echo "waiting for the cleanup script to exit..."
	    while kill -0 $CLEANUP_SCRIPT_PID > /dev/null 2>&1; do
		    sleep 1
	    done
    fi

# check if the container command is a singularity run command
# by checking if the first word in the command is "singularity"
# and the second word is "run"
elif [[ "${container_cmd[0]}" == "singularity" && "${container_cmd[1]}" == "run" ]]; then
    echo "singularity run strategy" >> debug.txt
    
    # run this container cmd
    echo "running job container..." >> debug.txt
    echo "container cmd: ${container_cmd[@]}" >> debug.txt
    # echo "container cmd: $container_cmd"
    eval "${container_cmd[@]}" > outputs/ospjob.log 2>&1

    # If we setup app proxying, run the cleanup script
    if [ "$setup_app_proxying" == "1" ]
    then
	    # cleanup
	    echo "Cleaning up..."
	    
	    # This will remove the revssh-client process
	    kill $REVSSH_CLIENT_PID
    fi
else
    echo "Invalid container command!"
fi
