#!/bin/bash
IS_DOCKER=$1
JOB_ID=$2

# If IS_DOCKER is 1, then we're running in a docker container
if [ "$IS_DOCKER" == "1" ]
then
	echo "Waiting for job ${JOB_ID} to finish..."
	docker wait ${JOB_ID}

	echo "Cleaning up..."


	# This will remove the revssh-client container
	docker rm -f revssh-${JOB_ID} || true
	docker network rm ${JOB_ID}

	echo "Job finished."
else
	JOB_PID_FILE=$3
	REVSSH_CLIENT_PID=$4
	echo "job id: $JOB_ID, job pid file: $JOB_PID_FILE, revssh client pid: $REVSSH_CLIENT_PID"                  

	# wait until the pid file is created
	echo "waiting for singularity pid file"
	while [ ! -f "$JOB_PID_FILE" ]; do
		sleep 1
	done

	# read the pid from the pid file
	JOB_PID=`cat $JOB_PID_FILE`
	echo "singularity pid: $JOB_PID"


	echo "Waiting for job ${JOB_ID} to finish..."

	# Check if the process ID is valid
	if ! kill -0 $JOB_PID > /dev/null 2>&1; then
		echo "Invalid process ID"
		exit 1
	fi

	# Wait for the process to exit
	while kill -0 $JOB_PID > /dev/null 2>&1; do
		sleep 1
	done

	echo "Cleaning up..."

	# This will remove the revssh-client process
	kill $REVSSH_CLIENT_PID

	echo "Job finished."   
fi
