#!/bin/bash
# Cleanup script for killing an executable job process and its shadow job container when either one is killed
watch_exec_pid=$1
exec_job_pid=$2
JOB_ID=$3
IS_DOCKER=$4

echo "Watch exec pid: $watch_exec_pid, exec job pid: $exec_job_pid, job id: $JOB_ID, is docker: $IS_DOCKER"

if [ "$watch_exec_pid" == "1" ]
then
	# Wait until the exec_job_pid process is killed
	# Once it's killed, delete the container "<job id>"

	# First, check if the process ID is valid
	if ! kill -0 $exec_job_pid > /dev/null 2>&1; then
		# Kill the container
		echo "Cleaning up shadow container..."
		if [ "$IS_DOCKER" == "1" ]
		then
			docker stop $JOB_ID || true
		else
			singularity instance stop $JOB_ID || true
		fi
		exit 0
	fi

	# Wait for the process to exit
	while kill -0 $exec_job_pid > /dev/null 2>&1; do
		sleep 1
	done

	# If IS_DOCKER is 1, then we're running in a docker container
	echo "Cleaning up shadow container..."
	if [ "$IS_DOCKER" == "1" ]
	then
		docker stop $JOB_ID || true
	else
		singularity instance stop $JOB_ID || true
	fi
else
	# Wait until the container "<job id>" is killed
	# Once it's killed, kill the exec_job_pid process

	# First, check if the container exists
	if [ "$IS_DOCKER" == "1" ]
	then
		if ! docker ps -a | grep $JOB_ID > /dev/null 2>&1; then
			# invalid container id
			echo "Cleaning up exec job process..."
			kill -9 $exec_job_pid
			exit 0
		fi
	else
		if ! singularity instance list | grep $JOB_ID > /dev/null 2>&1; then
			# invalid container id
			echo "Cleaning up exec job process..."
			kill -9 $exec_job_pid
			exit 0
		fi
	fi

	# Wait for the container to exit
	if [ "$IS_DOCKER" == "1" ]
	then
		docker wait $JOB_ID > /dev/null 2>&1
	else
		while singularity instance list | grep $JOB_ID > /dev/null 2>&1; do
			sleep 1
		done
	fi

	# Kill the exec_job_pid process
	echo "Cleaning up exec job process..."
	kill -9 $exec_job_pid
fi